# Cascader class reduces sequences of quantum gates for given gate-set and generates
# a multiplication table of all possible reductions
import timeit

import numpy as np
import time

from qworder.rules import Rules, Word


class Cascader(object):
    base_gates = {
        'I': np.array(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        'X': np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]]),
        'Y': np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]]),
        'Z': np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]]),
        'H': np.array([[0, 0, 1], [0, -1, 0], [1, 0, 0]]),
        'T': np.array(
            [[np.cos(np.pi / 4), -np.sin(np.pi / 4), 0], [np.sin(np.pi / 4), np.cos(np.pi / 4), 0], [0, 0, 1]]),
        'R': np.array(
            [[np.cos(np.pi / 4), np.sin(np.pi / 4), 0], [- np.sin(np.pi / 4), np.cos(np.pi / 4), 0], [0, 0, 1]])
    }

    def __init__(self, rules_path: str = ""):
        if len(rules_path):
            self.rules = Rules(rules_path)
        else:
            self.rules = Rules()

    def cascade_word(self, word: Word) -> Word:
        index = 0
        while index < len(word.word) - 1 and len(word.word) > 1:
            sub = word.word[index: index + 2]
            replacement = self._check_replace(sub)
            if replacement and replacement[0] != sub:
                word.replace(sub, replacement)
                index = max(0, index - 1)
            else:
                sub = word.word[index: index + 3]
                replacement = self._check_replace(sub)
                if replacement and replacement[0] != sub:
                    word.replace(sub, replacement)
                    index = max(0, index - 1)
                else:
                    index += 1

        self.rules.write_rules()
        return word

    def _check_replace(self, sub):
        if sub in self.rules and sub != self.rules[sub]:
            return self.rules[sub]
        replacement = self._check_product(sub)
        if replacement.word:
            self.rules[sub] = replacement
            return replacement
        else:
            self.rules[sub] = Word(sub, True)
            return False

    def _check_product(self, prod: str) -> Word:
        g = [self.base_gates[p] for p in prod]
        for letter in self.base_gates:
            temp = g[0]
            for i in range(1, len(g)):
                temp = np.matmul(temp, g[i])
            bg = self.base_gates[letter]
            if np.allclose(temp, bg):
                return Word(letter, True)
            elif np.allclose(temp, -bg):
                return Word(letter, False)
        return Word("", False)


if __name__ == '__main__':
    start_time = time.time()
    w = Word("HXYYXZXH", True)
    c = Cascader()
    print(w)
    print(c.cascade_word(w))
    print("--- %s seconds ---" % (time.time() - start_time))



import csv
import os

from qworder import config


class Rules(dict):

    def __init__(self, rules_path: str = ""):
        self._rules_path = rules_path if rules_path else config.PATH
        super().__init__(self._load_existing_rules())

    def rules(self):
        return self.rules

    def _load_existing_rules(self):
        rules = {}
        if os.path.isfile(self._rules_path):
            with open(self._rules_path) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                for row in csv_reader:
                    rules[row[0]] = [row[1], bool(row[2])]
        return rules

    def write_rules(self):
        output_file = open(self._rules_path, "w")
        for rule in self:
            output_file.write(rule + "," + self[rule][0] + "," + str(self[rule][1]) + "\n")
        output_file.close()


class Word(list):

    def __init__(self, word: str, sign: bool):
        super().__init__([word, sign])

    def __eq__(self, other):
        return self.word == other.word and self.sign == other.sign

    def replace(self, old: str, new: list):
        self[0] = self[0].replace(old, new[0])
        self[1] = self[1] == new[1]

    @property
    def word(self):
        return self[0]

    @property
    def sign(self):
        return self[1]



from distutils.core import setup
setup(
  name = 'qworder',
  packages = ['qworder'],   # Chose the same as "name"
  version = '0.111',
  license='gpl-3.0',
  description = 'QWorder simplifies strings representing sequences of quantum gates',
  author = 'Jakub Korsak',
  author_email = 'jakub.korsak@protonmail.com',
  url = 'https://github.com/korsakjakub/qworder',
  download_url = 'https://github.com/korsakjakub/QWorder/archive/refs/tags/0.1.tar.gz',
  keywords = ['QUANTUM', 'COMPUTING', 'GATESTRINGS', 'SIMPLIFIER', 'PAULI', 'MATRIX', 'HADAMARD'],
  install_requires=[
          'numpy',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
  ],
)



# Generate all possible words of given length
from typing import Dict, List


class WordGenerator:

    def __init__(self, input_set: list, length: int):
        self.output = []
        self.input_set = input_set
        self.length = length

    def generate_words(self) -> list:
        print("Generating words for length " + str(self.length))
        self.__generate_words_rec("", self.length)
        self._remove_unnecessary(self.length)
        return self.output

    def __generate_words_rec(self, word: str, length: int) -> None:
        if length == 0:
            self.output.append(word)
            return
        for i in range(len(self.input_set)):
            self.__generate_words_rec(word + self.input_set[i], length - 1)

    def generate_words_shorter_than(self) -> List[str]:
        for i in range(self.length):
            self.__generate_words_rec("", i + 1)
            self._remove_unnecessary(i + 1)
        return self.output

    def _remove_unnecessary(self, length: int = 0) -> None:
        if length == 1:
            return
        if length == 0:
            length = self.length
        for letter in self.input_set:
            self.output.remove(letter * length)

    def get_words_dictionary(self) -> Dict[int, list]:
        words = {}
        length = self.length
        for k in range(1, length + 1):
            self.length = k
            words[k] = self.generate_words()
            self.output = []
        return words
